/*
	XmBlackBerry, Copyright (C) 2008-2009  Rick Scott <rwscott@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include <linux/module.h>	/* Needed by all modules */
#include <linux/kernel.h>	/* Needed for KERN_INFO */
#include <linux/usb.h>
#include <linux/string.h>
#include <linux/tty.h>

#include "desktop.h"
#include "crypto.h"
#include "util.h"

static struct tty_driver *desktop_driver = NULL;

static unsigned char serial_header[] = {0xd9, 0xae, 0xfb};
static unsigned char serial_footer[] = {0xbf, 0xea, 0x9d};

static void
process_buffer(struct blackberry_mode *desktop)
{
struct blackberry_device *blackberry_device = container_of(desktop, struct blackberry_device, desktop);

    if (debug >= 9)
    {
	info("%s(%d) - %08x num_buffer = %i", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    desktop ? desktop->num_buffer : -1);
    }

    if (blackberry_device)
    {
	if (desktop->num_buffer > sizeof(serial_header))
	{
	    if (memcmp(&desktop->buffer[0], serial_header, sizeof(serial_header)) == 0)
	    {
		switch (desktop->buffer[sizeof(serial_header)])
		{
		case 0xcf:
		    {
		    unsigned char handshake_response[] = { 0xce };
		    int consumed = sizeof(serial_header) + 1;

			if (debug >= 2)
			{
			    info("%s(%d) - %08x", __FUNCTION__, __LINE__,
				blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
			    dump_hex("<-- desktop", desktop->buffer, desktop->num_buffer);
			}
			tty_send(desktop->tty, serial_header, sizeof(serial_header));
			tty_send(desktop->tty, handshake_response, sizeof(handshake_response));
			memmove(&desktop->buffer[0], &desktop->buffer[consumed], desktop->num_buffer - consumed);
			desktop->num_buffer -= consumed;
		    }
		    break;
		case 0x60:
		case 0x40:
		    if (desktop->num_buffer > sizeof(serial_header) + 3)
		    {
		    int size = 0;

			size = (size << 8) + desktop->buffer[sizeof(serial_header) + 2];
			size = (size << 8) + desktop->buffer[sizeof(serial_header) + 3];
			if (desktop->num_buffer >= sizeof(serial_header) + 1 + 2 + size + 2 + sizeof(serial_footer))
			{
			int consumed = sizeof(serial_header) + 1 + 2 + size + 2 + sizeof(serial_footer);

			    if (debug >= 2)
			    {
				info("%s(%d) - %08x", __FUNCTION__, __LINE__,
				    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
				dump_hex("<-- desktop", desktop->buffer, desktop->num_buffer);
			    }
			    if (desktop->num_buffer > sizeof(serial_header) + 4 + 2 + sizeof(serial_footer) &&
			        desktop->buffer[sizeof(serial_header) + 1] == 0x00 &&
			        desktop->buffer[sizeof(serial_header) + 4] == 'P' &&
			        desktop->buffer[sizeof(serial_header) + 1 + 2 + size] == 0x00)
			    {
				/*
				info("%s(%d) - %08x send password \"%s\" index = %i", __FUNCTION__, __LINE__,
				    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
				    &desktop->buffer[sizeof(serial_header) + 5],
				    desktop->tty->index);
				dump_hex("<-- desktop", desktop->buffer, desktop->num_buffer);
				*/
				if (blackberry_sha1_avail())
				{
				    if (device_map[desktop->tty->index].password_hash)
				    {
					blackberry_sha1_sum(&desktop->buffer[sizeof(serial_header) + 5],
					    strlen(&desktop->buffer[sizeof(serial_header) + 5]),
					    &device_map[desktop->tty->index].password_hash[4]);
					/* Don't leave the unhashed password laying around */
					memset(&desktop->buffer[sizeof(serial_header) + 5], 0, strlen(&desktop->buffer[sizeof(serial_header) + 5]));
					send_password(desktop, GFP_ATOMIC);
				    }
				    else
				    {
					err("%s(%d) - %08x", __FUNCTION__, __LINE__,
					    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
				    }
				}
				else
				{
				    err("%s(%d) - %08x", __FUNCTION__, __LINE__,
					blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
				}

			    }
			    else
			    {
				memcpy(&desktop->buffer[sizeof(serial_header) + 2], &desktop->buffer[sizeof(serial_header) + 0], 2);
				blackberry_write(blackberry_device,
				    desktop->socket,
				    &desktop->buffer[sizeof(serial_header) + 2],
				    consumed - sizeof(serial_header) - 2 - 2 - 2, GFP_ATOMIC);
			    }
			    memmove(&desktop->buffer[0], &desktop->buffer[consumed], desktop->num_buffer - consumed);
			    desktop->num_buffer -= consumed;
			}
		    }
		    break;
		case 0x41:
		    if (desktop->num_buffer >= sizeof(serial_header) + 1 + 2 + 0 + 2 + sizeof(serial_footer))
		    {
		    int consumed = sizeof(serial_header) + 1 + 2 + 0 + 2 + sizeof(serial_footer);

			if (debug >= 2)
			{
			    info("%s(%d) - %08x", __FUNCTION__, __LINE__,
				blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
			    dump_hex("<-- desktop", desktop->buffer, desktop->num_buffer);
			}
			blackberry_write(blackberry_device, desktop->socket, &desktop->buffer[sizeof(serial_header) + 0], 3, GFP_ATOMIC);
			memmove(&desktop->buffer[0], &desktop->buffer[consumed], desktop->num_buffer - consumed);
			desktop->num_buffer -= consumed;
		    }
		    break;
		default:
		    err("%s(%d) - %08x", __FUNCTION__, __LINE__,
			blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
		    dump_hex("<-- desktop", desktop->buffer, desktop->num_buffer);
		    desktop->num_buffer = 0;
		    break;
		}
	    }
	    else if (memcmp(&desktop->buffer[0], serial_footer, sizeof(serial_footer)) == 0)
	    {
		memmove(&desktop->buffer[0], &desktop->buffer[sizeof(serial_footer)], desktop->num_buffer - sizeof(serial_footer));
		desktop->num_buffer -= sizeof(serial_footer);
	    }
	    else
	    {
		/*
		err("%s(%d) - %08x", __FUNCTION__, __LINE__,
		    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
		dump_hex("<-- desktop", desktop->buffer, desktop->num_buffer);
		*/
		memmove(&desktop->buffer[0], &desktop->buffer[1], desktop->num_buffer - 1);
		desktop->num_buffer--;
	    }
	}
	else
	{
	    if (desktop->num_buffer > 0 &&
	        desktop->buffer[0] == 0x06)
	    {
		memmove(&desktop->buffer[0], &desktop->buffer[1], desktop->num_buffer - 1);
		desktop->num_buffer--;
	    }
	    else if ((memcmp(&desktop->buffer[0], serial_header, desktop->num_buffer) != 0) &&
	             (memcmp(&desktop->buffer[0], serial_footer, desktop->num_buffer) != 0))
	    {
		memmove(&desktop->buffer[0], &desktop->buffer[1], desktop->num_buffer - 1);
		desktop->num_buffer--;
	    }
	}
    }
    else
    {
	err("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
    }
}

static int
desktop_open(struct tty_struct *tty, struct file *file)
{
int ret = 0;
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;
struct blackberry_mode *mode = blackberry_device ? &blackberry_device->desktop : NULL;

    ret = mode_open(mode, tty, file);
    return(ret);
}

/**
	Update the crc with the next byte
*/
static u_int16_t
crc16(u_int16_t crcval, unsigned char cval)
{
u_int16_t poly;
int i;

    poly = 0x8408;
    crcval ^= cval;
    for (i = 0; i < 8; i++)
    {
    	crcval = crcval & 1 ?  (crcval >> 1) ^ poly : crcval >> 1;
    }
    return(crcval);
}

/**
	Compute the crc used by the serial protocol
*/
static u_int16_t
checksum(unsigned char *data, int len)
{
u_int16_t crc = 0xae9e;
int i;

    for (i = 0; i < len; i++)
    {
    	crc = crc16(crc, data[i]);
    }
    crc ^= 0x07f0;
    return(crc);
}

static int
desktop_send_packet(struct blackberry_mode *desktop, const void *data, int usb_len)
{
struct blackberry_device *blackberry_device = (desktop && desktop->tty) ? device_map[desktop->tty->index].blackberry_device : NULL;
int ret = 0;
const unsigned char *usb_data = data;
unsigned char *serial_data;
int serial_len;
int extra = 0;

    if (debug >= 3)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
	dump_hex("--> desktop", usb_data, usb_len);
    }
    if (blackberry_device)
    {
    unsigned char special[] = { 0x40, 0x00, 0x00 };

	if (usb_len >= sizeof(special))
	{
	    if (memcmp(usb_data, special, sizeof(special)) == 0)
	    {
		extra = 5;
	    }
	}

	serial_len = usb_len + sizeof(serial_header) + 2 + sizeof(serial_footer) + 2 + extra;
	serial_data = kmalloc(serial_len, GFP_KERNEL);
	if (serial_data)
	{
	u_int16_t sum;

	    memcpy(&serial_data[0], serial_header, sizeof(serial_header));
	    if (usb_len >= 3 && usb_data[0] == 0x41)
	    {
		serial_len = usb_len + sizeof(serial_header) + 2 + sizeof(serial_footer);
		memcpy(&serial_data[sizeof(serial_header)], &usb_data[0], 3);
		sum = checksum(&serial_data[sizeof(serial_header)], usb_len + 2);
		serial_data[sizeof(serial_header) + 3 + 0] = (sum >> 0) & 0xff;
		serial_data[sizeof(serial_header) + 3 + 1] = (sum >> 8) & 0xff;
		memcpy(&serial_data[sizeof(serial_header) + 3 + 2], serial_footer, sizeof(serial_footer));
		ret = tty_send(blackberry_device->desktop.tty, serial_data, serial_len);
	    }
	    else if (usb_len >= 3)
	    {
		memcpy(&serial_data[sizeof(serial_header)], &usb_data[0], 2);
		serial_data[sizeof(serial_header) + 2] = ((usb_len - 2 + extra) >> 8) & 0xff;
		serial_data[sizeof(serial_header) + 3] = ((usb_len - 2 + extra) >> 0) & 0xff;
		memcpy(&serial_data[sizeof(serial_header) + 4], &usb_data[2], usb_len - 2);

		if (extra >= 5)
		{
		    serial_data[sizeof(serial_header) + 4 + usb_len - 2 + 0] = (0) & 0xff;
		    serial_data[sizeof(serial_header) + 4 + usb_len - 2 + 1] = ((blackberry_device->pin) >> 24) & 0xff;
		    serial_data[sizeof(serial_header) + 4 + usb_len - 2 + 2] = ((blackberry_device->pin) >> 16) & 0xff;
		    serial_data[sizeof(serial_header) + 4 + usb_len - 2 + 3] = ((blackberry_device->pin) >>  8) & 0xff;
		    serial_data[sizeof(serial_header) + 4 + usb_len - 2 + 4] = ((blackberry_device->pin) >>  0) & 0xff;
		}

		sum = checksum(&serial_data[sizeof(serial_header)], usb_len + 2 + extra);
		serial_data[sizeof(serial_header) + 4 + usb_len - 2 + 0 + extra] = (sum >> 0) & 0xff;
		serial_data[sizeof(serial_header) + 4 + usb_len - 2 + 1 + extra] = (sum >> 8) & 0xff;
		memcpy(&serial_data[sizeof(serial_header) + 4 + usb_len - 2 + 2 + extra], serial_footer, sizeof(serial_footer));
		ret = tty_send(blackberry_device->desktop.tty, serial_data, serial_len);
	    }
	    else
	    {
		ret = -EINVAL;
		err("%s(%d) - %08x ENOMEM", __FUNCTION__, __LINE__,
		    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
		dump_hex("--> desktop", usb_data, usb_len);
	    }
	    kfree(serial_data);
	}
	else
	{
	    ret = -ENOMEM;
	    err("%s(%d) - %08x ENOMEM", __FUNCTION__, __LINE__,
		blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
	}
    }
    else
    {
	ret = -ENODEV;
	err("%s(%d) - %08x ENODEV", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
    }
    return(ret);
}

int
desktop_register_device(struct blackberry_mode *desktop, int index)
{
int ret = 0;
struct blackberry_device *blackberry_device = device_map[index].blackberry_device;

    if (debug >= 1)
    {
	info("%s(%d) - %08x /dev/%s%i", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    desktop_driver->name,
	    index);
    }
    desktop->mode = "RIM Desktop";
    desktop->write = process_buffer;
    desktop->send_packet = desktop_send_packet;
    ret = mode_register_device(desktop_driver, desktop, index);
    return(ret);
}

int
desktop_deregister_device(struct blackberry_mode *desktop, int index)
{
int ret = 0;
struct blackberry_device *blackberry_device = device_map[index].blackberry_device;

    if (debug >= 1)
    {
	info("%s(%d) - %08x /dev/%s%i", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    desktop_driver->name,
	    index);
    }
    ret = mode_deregister_device(desktop_driver, desktop, index);
    return(ret);
}

int
desktop_register(void)
{
int ret = 0;

    if (debug >= 8)
    {
	info("%s(%d)", __FUNCTION__, __LINE__);
    }

    desktop_driver = mode_register("bbdesktop", desktop_open);
    if (desktop_driver)
    {
    	desktop_driver->driver_name = "Blackberry Desktop";
    }
    else
    {
    	ret = -ENOMEM;
	err("%s(%d)", __FUNCTION__, __LINE__);
    }
    return(ret);
}

int
desktop_deregister(void)
{
int ret = 0;

    if (debug >= 8)
    {
	info("%s(%d)", __FUNCTION__, __LINE__);
    }
    ret = mode_deregister(desktop_driver);
    return(ret);
}

