/*
	XmBlackBerry, Copyright (C) 2008-2009  Rick Scott <rwscott@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    Copyright 2008, Chris Frey, NetDirect (http://www.netdirect.ca/), GPL v2
    2008/11 - Now compiles on 2.6.26.5: Changes to URB callback mechanism and
	      work queue API.  Also uses new sha1 API.
*/

#include "config.h"

#include <linux/module.h>	/* Needed by all modules */
#include <linux/kernel.h>	/* Needed for KERN_INFO */
#include <linux/init.h>
#include <linux/usb.h>
#include <linux/string.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/mm.h>
#ifdef HAVE_LOCK_KERNEL
#include <linux/smp_lock.h>
#endif
#include <asm/uaccess.h>

#include "blackberry.h"
#include "desktop.h"
#include "charge.h"
#include "util.h"
#include "modem.h"
#include "loader.h"
#include "ftp.h"
#include "caltask.h"
#include "jvmdebug.h"
#include "bootload.h"
#include "crypto.h"

MODULE_LICENSE("GPL v2");
MODULE_AUTHOR(PACKAGE_BUGREPORT);
MODULE_DESCRIPTION("Device driver for the RIM Blackberry");
MODULE_VERSION(PACKAGE_VERSION);

int debug = 0;
module_param(debug, int, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
MODULE_PARM_DESC(debug, "Debug verbosity level. 0 = none");

static char *password = NULL;
module_param(password, charp, S_IRUSR);
MODULE_PARM_DESC(password, "Password to use for next plugged in device");

int num_devices = 2;
module_param(num_devices, int, S_IRUSR | S_IRGRP | S_IROTH);
MODULE_PARM_DESC(num_devices, "Number of Blackberrys (default 2)");

struct device_map *device_map = NULL;

static int probe(struct usb_interface *intf, const struct usb_device_id *id);
static void disconnect(struct usb_interface *intf);

static struct usb_device_id blackberry_table[] = {
    {USB_DEVICE(0x0fca, 0x0001)},
    {USB_DEVICE(0x0fca, 0x0004)},
    {USB_DEVICE(0x0fca, 0x0006)},
    {USB_DEVICE(0x0fca, 0x8004)},
    {}
};

static struct usb_driver blackberry_driver = {
    .name = "blackberry",
    .probe = probe,
    .disconnect = disconnect,
    .id_table = blackberry_table,
};

static int blackberry_read(struct blackberry_device *blackberry_device);

struct workqueue_struct *work_queue = NULL;

static void mode_close(struct tty_struct *tty, struct file *file);
#ifdef FROM_USER_IN_TTY_OPS
static int mode_write(struct tty_struct *tty, int from_user, const unsigned char *buf, int count);
#else
static int mode_write(struct tty_struct *tty, const unsigned char *buf, int count);
#endif
static int mode_write_room(struct tty_struct *tty);
static int mode_chars_in_buffer(struct tty_struct *tty);
#ifdef FILE_IN_TTY_OPS_TIOCM
static int mode_tiocmget(struct tty_struct *tty, struct file *file);
static int mode_tiocmset(struct tty_struct *tty, struct file *file, unsigned int set, unsigned int clear);
#else
static int mode_tiocmget(struct tty_struct *tty);
static int mode_tiocmset(struct tty_struct *tty, unsigned int set, unsigned int clear);
#endif

static const struct tty_operations generic_ops = {
    .close = mode_close,
    .write = mode_write,
    .write_room = mode_write_room,
    .chars_in_buffer = mode_chars_in_buffer,
    .tiocmget = mode_tiocmget,
    .tiocmset = mode_tiocmset,
};

int
mode_open(struct blackberry_mode *mode, struct tty_struct *tty, struct file *file)
{
int ret = 0;
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;

    if (debug >= 7)
    {
	info("%s(%d) - %08x /dev/%s driver_data = 0x%p open_count = %i %sconnected", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    (tty && tty->name) ? tty->name : "None",
	    tty ? tty->driver_data : NULL,
	    mode ? mode->open_count : -1,
	    mode ? (mode->connected ? "" : "not ") : "not ");
    }

    if (mode && (blackberry_device = blackberry_get(blackberry_device)))
    {
	if (down_interruptible(&mode->sem))
	{
	    ret = -ERESTARTSYS;
	}
	else
	{
	    if (tty->driver_data == NULL)
	    {
		blackberry_get(blackberry_device);
		tty->driver_data = mode;
		mode->tty = tty;
	    }
	    ++mode->open_count;
	    if (mode->open_count == 1)
	    {
		tty->low_latency = 1;
		if (mode->socket < 0)
		{
		    mode->socket = -2;
		    select_mode(blackberry_device, mode->mode);
		    wait_event_interruptible(mode->open_wait, (mode->socket != -2));
		}
		if (mode->socket > 0)
		{
		    open_socket(blackberry_device, mode->socket);
		    if (file->f_flags & O_NONBLOCK)
		    {
		    }
		    else
		    {
			wait_event_interruptible(mode->open_wait, (mode->connected != 0));
			if (mode->connected != 1)
			{
			    ret = -ENODEV;
			    err("%s(%d) - %08x %s not available", __FUNCTION__, __LINE__,
				blackberry_device ? blackberry_device->pin: (u_int32_t)-1,
				mode->mode);
			    tty->driver_data = NULL;
			    mode->tty = NULL;
			    --mode->open_count;
			    blackberry_put(blackberry_device);
			}
		    }
		}
		else
		{
		    if (mode == &blackberry_device->bootload)
		    {
			if (device_map[tty->index].attempt_bootload == 0)
			{
			    device_map[tty->index].attempt_bootload = 1;
			    reset_device(blackberry_device);
			}
		    }
		    ret = -ENODEV;
		    err("%s(%d) - %08x \"%s\" not available", __FUNCTION__, __LINE__,
			blackberry_device ? blackberry_device->pin: (u_int32_t)-1,
			mode ? mode->mode : "NULL");
		    tty->driver_data = NULL;
		    mode->tty = NULL;
		    --mode->open_count;
		    blackberry_put(blackberry_device);
		}
	    }
	    up(&mode->sem);
	}
	blackberry_put(blackberry_device);
    }
    else
    {
	ret = -ENODEV;
	err("%s(%d) - %08x %s not available", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin: (u_int32_t)-1,
	    mode ? mode->mode : "NULL");
    }
    return(ret);
}

static void
mode_close(struct tty_struct *tty, struct file *file)
{
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;
struct blackberry_mode *mode = tty ? tty->driver_data : NULL;

    if (debug >= 7)
    {
	info("%s(%d) - %08x /dev/%s driver_data = 0x%p open_count = %i %sconnected", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    (tty && tty->name) ? tty->name : "None",
	    tty ? tty->driver_data : NULL,
	    mode ? mode->open_count : -1,
	    mode ? (mode->connected ? "" : "not ") : "not ");
    }

    if (mode && (blackberry_device = blackberry_get(blackberry_device)))
    {
	if (mode->socket > 0)
	{
	    down(&mode->sem);
	    if (mode->open_count > 0)
	    {
		--mode->open_count;
		if (mode->open_count <= 0)
		{
		    if (debug >= 6)
		    {
			info("%s(%d) - %08x /dev/%s Close %s socket 0x%04x", __FUNCTION__, __LINE__,
			    blackberry_device->pin,
			    tty->name,
			    mode->mode ? mode->mode : "NULL",
			    mode->socket);
		    }
		    close_socket(blackberry_device, mode->socket);
		    if (file->f_flags & O_NONBLOCK)
		    {
		    }
		    else
		    {
			wait_event_interruptible(mode->open_wait, (mode->connected != 1));
		    }
		    tty->driver_data = NULL;
		    mode->tty = NULL;
		    mode->num_buffer = 0;
		    blackberry_put(blackberry_device);
		}
	    }
	    else
	    {
		err("%s(%d) - %08x close /dev/%s without an open", __FUNCTION__, __LINE__,
		    blackberry_device->pin,
		    tty->name);
	    }
	    up(&mode->sem);
	}
	else
	{
	    err("%s(%d) - %08x %s not available", __FUNCTION__, __LINE__,
		blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
		mode ? mode->mode : "NULL");
	}
	blackberry_put(blackberry_device);
    }
    else
    {
	err("%s(%d) - %08x %s not available", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    mode ? mode->mode : "NULL");
    }
}

static int
#ifdef FROM_USER_IN_TTY_OPS
mode_write(struct tty_struct *tty, int from_user, const unsigned char *buf, int count)
#else
mode_write(struct tty_struct *tty, const unsigned char *buf, int count)
#endif
{
int written = 0;
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;
struct blackberry_mode *mode = tty ? tty->driver_data : NULL;

    if (mode && (blackberry_device = blackberry_get(blackberry_device)))
    {
	if (mode->open_count)
	{
	int i;

	    for (i = 0; (i < count) && sizeof(mode->buffer) - mode->num_buffer > 0; i++)
	    {
#ifdef FROM_USER_IN_TTY_OPS
		if (from_user)
		{
		    copy_from_user(&mode->buffer[mode->num_buffer], &buf[i], 1);
		}
		else
		{
		    mode->buffer[mode->num_buffer] = buf[i];
		}
#else
		mode->buffer[mode->num_buffer] = buf[i];
#endif
		mode->num_buffer++;
		if (mode->write)
		{
		    (*mode->write)(mode);
		}
	    }
	    if (mode->write == NULL)
	    {
		blackberry_write(blackberry_device, mode->socket, &mode->buffer[0], mode->num_buffer, GFP_ATOMIC);
		mode->num_buffer = 0;
	    }
	    if (sizeof(mode->buffer) - mode->num_buffer <= 0)
	    {
		written = -ENOMEM;
	    }
	    else
	    {
		written = i;
	    }
	}
	else
	{
	    written = -ENODEV;
	}
	blackberry_put(blackberry_device);
    }
    else
    {
	written = -ENODEV;
    }
    return(written);
}

static int
mode_write_room(struct tty_struct *tty)
{
int room = 0;
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;
struct blackberry_mode *mode = tty ? tty->driver_data : NULL;

    if (debug >= 9)
    {
	info("%s(%d) - %08x tty = %p max_packet = %i", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    tty,
	    blackberry_device ? blackberry_device->max_packet : -1);
    }

    if (mode)
    {
	room = mode ? sizeof(mode->buffer) - mode->num_buffer : -ENODEV;
    }
    return(room);
}

static int
mode_chars_in_buffer(struct tty_struct *tty)
{
int num_chars = 0;
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;
struct blackberry_mode *mode = tty ? tty->driver_data : NULL;

    if (debug >= 9)
    {
	info("%s(%d) - %08x tty = %p", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    tty);
    }
    num_chars = mode ? mode->num_buffer : -ENODEV;

    return(num_chars);
}

static int
#ifdef FILE_IN_TTY_OPS_TIOCM
mode_tiocmget(struct tty_struct *tty, struct file *file)
#else
mode_tiocmget(struct tty_struct *tty)
#endif
{
int ret = 0;
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;
struct blackberry_mode *mode = tty ? tty->driver_data : NULL;

    if (debug >= 9)
    {
	info("%s(%d) - %08x tty = %p", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    tty);
    }

    if (mode)
    {
	ret = ((mode->mcr & TIOCM_DTR)  ? TIOCM_DTR  : 0) | /* DTR is set */
	      ((mode->mcr & TIOCM_RTS)  ? TIOCM_RTS  : 0) | /* RTS is set */
	      ((mode->mcr & TIOCM_LOOP) ? TIOCM_LOOP : 0) | /* LOOP is set */
	      ((mode->msr & TIOCM_CTS)  ? TIOCM_CTS  : 0) | /* CTS is set */
	      ((mode->msr & TIOCM_CD)   ? TIOCM_CAR  : 0) | /* Carrier detect is set*/
	      ((mode->msr & TIOCM_RI)   ? TIOCM_RI   : 0) | /* Ring Indicator is set */
	      ((mode->msr & TIOCM_DSR)  ? TIOCM_DSR  : 0);  /* DSR is set */
    }
    else
    {
    	ret = -ENODEV;
    }

    return(ret);
}

static int
#ifdef FILE_IN_TTY_OPS_TIOCM
mode_tiocmset(struct tty_struct *tty, struct file *file, unsigned int set, unsigned int clear)
#else
mode_tiocmset(struct tty_struct *tty, unsigned int set, unsigned int clear)
#endif
{
int ret = 0;
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;
//struct blackberry_mode *mode = tty ? tty->driver_data : NULL;

    if (debug >= 8)
    {
	info("%s(%d) - %08x tty = %p set = 0x%08x clear = 0x%08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    tty, set, clear);
    }

    return(blackberry_device ? ret : -ENODEV);
}

int
mode_register_device(struct tty_driver *driver, struct blackberry_mode *mode, int index)
{
int ret = 0;
struct blackberry_device *blackberry_device = device_map[index].blackberry_device;

    if (debug >= 1)
    {
	info("%s(%d) - %08x /dev/%s%i", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    driver->name,
	    index);
    }
    if ((blackberry_device = blackberry_get(blackberry_device)))
    {
	init_waitqueue_head(&mode->open_wait);
	sema_init(&mode->sem, 1);
	mode->open_count = 0;
	mode->socket = -1;
	mode->connected = 0;
	mode->num_buffer = 0;
	mode->msr = TIOCM_CTS;
	mode->mcr = TIOCM_RTS | TIOCM_DTR;
	tty_register_device(driver, index, NULL);
    }
    else
    {
    	ret = -ENODEV;
	err("%s(%d)", __FUNCTION__, __LINE__);
    }
    return(ret);
}

int
mode_deregister_device(struct tty_driver *driver, struct blackberry_mode *mode, int index)
{
int ret = 0;
struct blackberry_device *blackberry_device = device_map[index].blackberry_device;

    if (debug >= 1)
    {
	info("%s(%d) - %08x /dev/%s%i", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    driver->name,
	    index);
    }
    if ((blackberry_device = blackberry_get(blackberry_device)))
    {
	mode->socket = -1;
	mode->connected = -1;
	wake_up_interruptible(&mode->open_wait);
	tty_unregister_device(driver, index);
	blackberry_put(blackberry_device);
	blackberry_put(blackberry_device);
    }
    else
    {
    	ret = -ENODEV;
	err("%s(%d)", __FUNCTION__, __LINE__);
    }
    return(ret);
}

struct tty_driver *
mode_register(const char *name, int  (*open)(struct tty_struct * tty, struct file * filp))
{
int ret = 0;
struct tty_driver *driver;
static struct tty_operations *mode_ops;

    if (debug >= 1)
    {
	info("%s(%d) - \"%s\"", __FUNCTION__, __LINE__,
	    name);
    }

    driver = alloc_tty_driver(num_devices);
    mode_ops = kmalloc(sizeof(struct tty_operations), GFP_KERNEL);
    if (driver && mode_ops)
    {
    	driver->owner = THIS_MODULE;
    	driver->name = name;
    	driver->flags = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW
#ifdef TTY_DRIVER_NO_DEVFS
    	| TTY_DRIVER_NO_DEVFS;
#else
    	| TTY_DRIVER_DYNAMIC_DEV;
#endif
    	driver->init_termios = tty_std_termios;
    	driver->init_termios.c_iflag = IGNBRK | IGNPAR;
    	driver->init_termios.c_oflag = 0;
    	driver->init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;
    	driver->init_termios.c_lflag = NOFLSH;
    	*mode_ops = generic_ops;
    	mode_ops->open = open;
    	tty_set_operations(driver, mode_ops);
    	ret = tty_register_driver(driver);
    	if (ret)
    	{
	    err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	    put_tty_driver(driver);
	    driver = NULL;
    	}
    }
    else
    {
	err("%s(%d) - mode_ops %p driver %p", __FUNCTION__, __LINE__,
	    mode_ops,
	    driver);
	if (mode_ops) kfree(mode_ops);
	if (driver)
	{
	    put_tty_driver(driver);
	    driver = NULL;
	}
    }
    return(driver);
}

int
mode_deregister(struct tty_driver *driver)
{
int ret = 0;

    if (debug >= 1)
    {
	info("%s(%d) - \"%s\"", __FUNCTION__, __LINE__,
	    driver->name);
    }
    tty_unregister_driver(driver);
#ifdef TTY_DRIVER_OPS_PTR
    kfree(driver->ops);
#endif
    put_tty_driver(driver);
    return(ret);
}

int
tty_send(struct tty_struct *tty, const unsigned char *buf, int count)
{
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;
int i = 0;

    if (debug >= 2)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
	dump_hex("--> tty", buf, count);
    }

    if (tty)
    {
#ifdef HAVE_TTY_FLIP_BUFFER
	for (i = 0; i < count; i++)
	{
	    if (tty->flip.count >= (TTY_FLIPBUF_SIZE / 2))
	    {
		tty_flip_buffer_push(tty);
	    }
	    tty_insert_flip_char(tty, buf[i], TTY_NORMAL);
	}
	tty_flip_buffer_push(tty);
#else
	for (i = 0; i < count; i++)
	{
	    if (tty_buffer_request_room(tty, 1))
	    {
		tty_insert_flip_char(tty, buf[i], TTY_NORMAL);
	    }
	}
	tty_schedule_flip(tty);
#endif
    }
    else
    {
    	i = -ENODEV;
    }
    return(i);
}

static void
#ifdef USB_COMPLETE_NEEDS_REGS
blackberry_write_callback(struct urb *urb, struct pt_regs *regs)
#else
blackberry_write_callback(struct urb *urb)
#endif
{
struct blackberry_device *blackberry_device = urb->context;

    if (debug >= 9)
    {
	info("%s(%d) - %08x status = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, urb->status);
    }

    usb_free_coherent(urb->dev,
    	urb->transfer_buffer_length, urb->transfer_buffer,
    	urb->transfer_dma);
    blackberry_put(blackberry_device);
}

int
blackberry_write(struct blackberry_device *blackberry_device, int socket, const void *input, int len, int mem_flags)
{
int written = 0;

    if (debug >= 9)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device->pin);
    }

    if (len > 0)
    {
    struct urb *urb;

	urb = usb_alloc_urb(0, mem_flags);
	if (urb)
	{
	char *data;

	    data = usb_alloc_coherent(blackberry_device->usb_device, len + 4, mem_flags, &urb->transfer_dma);
	    if (data)
	    {
	    int ret;
	    unsigned long flags;

		data[0] = (socket >> 0) & 0xff;
		data[1] = (socket >> 8) & 0xff;
		data[2] = ((len + 4) >> 0) & 0xff;
		data[3] = ((len + 4) >> 8) & 0xff;
		memcpy(&data[4], input, len);

		blackberry_get(blackberry_device);
		usb_fill_bulk_urb(urb, blackberry_device->usb_device,
		    usb_sndbulkpipe(blackberry_device->usb_device, blackberry_device->write_ep),
		    data, len + 4,
		    blackberry_write_callback, blackberry_device);
		urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;

		spin_lock_irqsave(&blackberry_device->urb_queue_lock, flags);
		{
		    if (socket == 0 || blackberry_device->last_urb == NULL)
		    {
			ret = usb_submit_urb(urb, mem_flags);
			if (ret == 0)
			{
			    if (debug >= 1)
			    {
				info("%s(%d) - %08x socket %i", __FUNCTION__, __LINE__,
				    blackberry_device->pin,
				    socket);
				dump_hex("--> usb", data, (len + 4));
			    }
			    written = len;
			    if (socket != 0)
			    {
				blackberry_device->last_urb = usb_get_urb(urb);
			    }
			}
			else
			{
			    written = ret;
			    usb_free_coherent(urb->dev,
			    	urb->transfer_buffer_length, urb->transfer_buffer,
			    	urb->transfer_dma);
			    blackberry_put(blackberry_device);
			}
		    }
		    else
		    {
			written = len;
			usb_get_urb(urb);
			list_add_tail(&urb->urb_list, &blackberry_device->urb_queue);
		    }
		}
		spin_unlock_irqrestore(&blackberry_device->urb_queue_lock, flags);
	    }
	    else
	    {
		written = -ENOMEM;
		err("%s(%d) - %08x ENOMEM", __FUNCTION__, __LINE__,
		    blackberry_device->pin);
	    }
	    usb_free_urb(urb);
	}
	else
	{
	    written = -ENOMEM;
	    err("%s(%d) - %08x ENOMEM", __FUNCTION__, __LINE__,
		blackberry_device->pin);
	}
    }
    else
    {
	err("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
    }
    return(written);
}

int
send_password(struct blackberry_mode *mode, int mem_flags)
{
int ret = 0;
struct blackberry_device *blackberry_device = (mode && mode->tty) ? device_map[mode->tty->index].blackberry_device : NULL;

    if (blackberry_device)
    {
    u8 *response = NULL;

	if (blackberry_sha1_avail())
	{
	    response = kmalloc(8 + blackberry_sha1_digestsize(), mem_flags);
	}
	if (response)
	{
	    blackberry_sha1_sum(&device_map[mode->tty->index].password_hash[0],
		4 + blackberry_sha1_digestsize(),
		&response[8]);
	    response[0] = 0x0f;
	    response[1] = (mode->socket >> 0) & 0xff;
	    response[2] = (mode->socket >> 8) & 0xff;
	    response[3] = blackberry_device->attempts_remaining + 1;
	    response[4] = 0x00;
	    response[5] = 0x00;
	    response[6] = blackberry_sha1_digestsize();
	    response[7] = 0x00;
	    blackberry_write(blackberry_device, 0, response, 8 + blackberry_sha1_digestsize(), GFP_ATOMIC);
	    kfree(response);
	}
	else
	{
	    ret = -ENOMEM;
	}
    }
    else
    {
	ret = -ENODEV;
    }
    return(ret);
}

static int
get_info(struct blackberry_device *blackberry_device, int what)
{
char cmd[] = { 0x05, 0xff, 0x00, 0x00, 0x00, 0x00, 0x02, 0x00 };
int written;

    if (debug >= 8)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device->pin);
    }

    cmd[6] = (what >> 0) & 0xff;
    cmd[7] = (what >> 8) & 0xff;
    written = blackberry_write(blackberry_device, 0, cmd, sizeof(cmd), GFP_KERNEL);
    if (written != sizeof(cmd))
    {
	err("%s(%d) - %08x written = %i sizeof(desc) = %zu", __FUNCTION__, __LINE__,
	    blackberry_device->pin, written, sizeof(cmd));
    }
    return(written);
}

static void
get_description(struct blackberry_device *blackberry_device)
{
    get_info(blackberry_device, 0x0002);
}

static void
get_pin(struct blackberry_device *blackberry_device)
{
    get_info(blackberry_device, 0x0004);
}

void
select_mode(struct blackberry_device *blackberry_device, const char *mode)
{
char packet[16 + 4];
int written = 0;

    if (debug >= 8)
    {
	info("%s(%d) - %08x mode = \"%s\"", __FUNCTION__, __LINE__,
	    blackberry_device->pin,
	    mode ? mode : "NULL");
    }

    if (mode && (strlen(mode) < sizeof(packet) - 5))
    {
	memset(packet, 0, sizeof(packet));
	packet[0] = 0x07;
	packet[1] = 0xff;
	packet[2] = 0x00;
	packet[3] = strlen(mode);
	memcpy(&packet[4], mode, strlen(mode) + 1);
	written = blackberry_write(blackberry_device, 0, packet, sizeof(packet), GFP_KERNEL);
	if (written != sizeof(packet))
	{
	    err("%s(%d) - %08x written = %i sizeof(packet) = %zu", __FUNCTION__, __LINE__,
		blackberry_device->pin, written, sizeof(packet));
	}
    }
    else
    {
	err("%s(%d) - %08x strlen(mode) = %zu", __FUNCTION__, __LINE__,
	    blackberry_device->pin,
	    mode ? strlen(mode) : -1);
    }
}

void
open_socket(struct blackberry_device *blackberry_device, int socket)
{
char cmd[] = { 0x0a, 0x00, 0x00 };
int written = 0;

    if (debug >= 8)
    {
	info("%s(%d) - %08x socket = 0x%04x", __FUNCTION__, __LINE__,
	    blackberry_device->pin, socket);
    }

    cmd[1] = (socket >> 0) & 0xff;
    cmd[2] = (socket >> 8) & 0xff;
    written = blackberry_write(blackberry_device, 0, cmd, sizeof(cmd), GFP_KERNEL);
    if (written != sizeof(cmd))
    {
	err("%s(%d) - %08x written = %i sizeof(cmd) = %zu", __FUNCTION__, __LINE__,
	    blackberry_device->pin, written, sizeof(cmd));
    }
}

void
close_socket(struct blackberry_device *blackberry_device, int socket)
{
char cmd[] = { 0x0b, 0x00, 0x00 };
int written = 0;

    if (debug >= 8)
    {
	info("%s(%d) - %08x socket = 0x%04x", __FUNCTION__, __LINE__,
	    blackberry_device->pin, socket);
    }

    cmd[1] = (socket >> 0) & 0xff;
    cmd[2] = (socket >> 8) & 0xff;
    written = blackberry_write(blackberry_device, 0, cmd, sizeof(cmd), GFP_KERNEL);
    if (written != sizeof(cmd))
    {
	err("%s(%d) - %08x written = %i sizeof(cmd) = %zu", __FUNCTION__, __LINE__,
	    blackberry_device->pin, written, sizeof(cmd));
    }
}

void
reset_device(struct blackberry_device *blackberry_device)
{
char cmd[] = { 0x03, 0xff, 0x00, 0x0a };
int written = 0;

    if (debug >= 8)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device->pin);
    }

    written = blackberry_write(blackberry_device, 0, cmd, sizeof(cmd), GFP_KERNEL);
    if (written != sizeof(cmd))
    {
	err("%s(%d) - %08x written = %i sizeof(cmd) = %zu", __FUNCTION__, __LINE__,
	    blackberry_device->pin, written, sizeof(cmd));
    }
}

static int
blackberry_register_device(struct blackberry_device *blackberry_device)
{
int ret = 0;
int i;

    if (debug >= 9)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device->pin);
    }

    for (i = 0; i < num_devices; i++)
    {
    	if (blackberry_device->pin == device_map[i].pin)
    	{
	    if (debug >= 1)
	    {
		info("%s(%d) - %08x Seen it before index = %i", __FUNCTION__, __LINE__,
		    blackberry_device->pin, i);
	    }
	    break;
    	}
    }
    if (i >= num_devices)
    {
	for (i = 0; i < num_devices; i++)
    	{
	    if (device_map[i].pin == (u_int32_t)-1)
	    {
		break;
	    }
    	}
    }
    if (i < num_devices)
    {
	info("%s(%d) - %08x device %i attempt_bootload %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin,
	    i,
	    device_map[i].attempt_bootload);
	blackberry_get(blackberry_device);
	device_map[i].blackberry_device = blackberry_device;
	device_map[i].pin = blackberry_device->pin;
	if (blackberry_sha1_avail() && device_map[i].password_hash == NULL && password != NULL)
	{
	    device_map[i].password_hash = kmalloc(4 + blackberry_sha1_digestsize(), GFP_KERNEL);
	    if (device_map[i].password_hash)
	    {
		blackberry_sha1_sum(password,
		    strlen(password),
		    &device_map[i].password_hash[4]);
	    }
	    else
	    {
		err("%s(%d) - %08x", __FUNCTION__, __LINE__,
		    blackberry_device->pin);
	    }
	}
	else
	{
	    if (password != NULL)
	    {
		err("%s(%d) - %08x sha1 %sAvailable hash %p password %p", __FUNCTION__, __LINE__,
		    blackberry_device->pin,
		    blackberry_sha1_avail() ? "" : "NOT ",
		    device_map[i].password_hash,
		    password);
	    }
	}
	desktop_register_device(&blackberry_device->desktop, i);
	modem_register_device(&blackberry_device->modem, i);
	loader_register_device(&blackberry_device->loader, i);
	ftp_register_device(&blackberry_device->ftp, i);
	caltask_register_device(&blackberry_device->caltask, i);
	jvmdebug_register_device(&blackberry_device->jvmdebug, i);
	bootload_register_device(&blackberry_device->bootload, i);
    }
    else
    {
    	ret = -ENODEV;
	err("%s(%d) - %08x Need to increase num_devices module parameter. Currently %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, num_devices);
    }
    return(ret);
}

static void
blackberry_unregister_device(struct blackberry_device *blackberry_device)
{
int i;

    if (debug >= 8)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
    }

    if (blackberry_device)
    {
	for (i = 0; i < num_devices; i++)
	{
	    if (blackberry_device->pin != (u_int32_t)-1 && blackberry_device->pin == device_map[i].pin)
	    {
		info("%s(%d) - %08x device %i", __FUNCTION__, __LINE__,
		    blackberry_device->pin, i);
		bootload_deregister_device(&blackberry_device->bootload, i);
		jvmdebug_deregister_device(&blackberry_device->jvmdebug, i);
		caltask_deregister_device(&blackberry_device->caltask, i);
		ftp_deregister_device(&blackberry_device->ftp, i);
		loader_deregister_device(&blackberry_device->loader, i);
		modem_deregister_device(&blackberry_device->modem, i);
		desktop_deregister_device(&blackberry_device->desktop, i);
		device_map[i].blackberry_device = NULL;
		blackberry_put(blackberry_device);
		break;
	    }
	}
	if (i >= num_devices)
	{
	    err("%s(%d) - %08x not registered", __FUNCTION__, __LINE__,
		blackberry_device->pin);
	}
    }
    else
    {
	err("%s(%d)", __FUNCTION__, __LINE__);
    }
}

static int
parse_pin(struct blackberry_device *blackberry_device, int len, const u_int8_t *data)
{
int ret = 0;

    if (debug >= 8)
    {
	info("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
    }

    if (len >= 8)
    {
    	blackberry_device->pin = 0;
    	blackberry_device->pin = (blackberry_device->pin << 8) + data[7];
    	blackberry_device->pin = (blackberry_device->pin << 8) + data[6];
    	blackberry_device->pin = (blackberry_device->pin << 8) + data[5];
    	blackberry_device->pin = (blackberry_device->pin << 8) + data[4];
    	ret = blackberry_register_device(blackberry_device);
    }
    else
    {
	ret = -EINVAL;
	err("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
	dump_hex("<-- usb", data, len);
    }
    return(ret);
}

static int
validate_mode(struct blackberry_mode *mode, int socket, const u_int8_t *data)
{
int ret = 0;
struct blackberry_device *blackberry_device = (mode && mode->tty) ? device_map[mode->tty->index].blackberry_device : NULL;

    if (strlen(mode->mode) == data[0] &&
	strncmp(mode->mode, &data[1], data[0]) == 0)
    {
	if (socket != 0x00ff)
	{
	    if (debug >= 8)
	    {
		info("%s(%d) - %08x mode \"%s\" on socket 0x%04x", __FUNCTION__, __LINE__,
		    blackberry_device? blackberry_device->pin : (u_int32_t)-1,
		    &data[1], socket);
	    }
	    mode->socket = socket;
	}
	else
	{
	    err("%s(%d) - %08x Mode \"%s\" not available", __FUNCTION__, __LINE__,
		blackberry_device? blackberry_device->pin : (u_int32_t)-1,
		&data[1]);
	    mode->socket = -1;
	    ret = -ENOSYS;
	}
	wake_up_interruptible_sync(&mode->open_wait);
    }
    else
    {
	ret = -ENOSYS;
    }
    return(ret);
}

static int
parse_mode_select(struct blackberry_device *blackberry_device, int len, const u_int8_t *data)
{
int ret = 0;

    if (debug >= 9)
    {
	info("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
    }

    if (len > 3)
    {
    int socket = 0;

	socket = (socket << 8) + data[1];
	socket = (socket << 8) + data[0];
	if ((ret = validate_mode(&blackberry_device->desktop, socket, &data[2])) == 0)
	{
	}
	else if ((ret = validate_mode(&blackberry_device->loader, socket, &data[2])) == 0)
	{
	}
	else if ((ret = validate_mode(&blackberry_device->ftp, socket, &data[2])) == 0)
	{
	}
	else if ((ret = validate_mode(&blackberry_device->caltask, socket, &data[2])) == 0)
	{
	}
	else if ((ret = validate_mode(&blackberry_device->jvmdebug, socket, &data[2])) == 0)
	{
	}
	else if ((ret = validate_mode(&blackberry_device->bootload, socket, &data[2])) == 0)
	{
	}
	else
	{
	    //ret = -ENOSYS;
	    err("%s(%d) - %08x Mode \"%s\" not available on 0x%04x", __FUNCTION__, __LINE__,
		blackberry_device->pin, &data[3], socket);
	    dump_hex("<-- usb", data, len);
	}
    }
    else
    {
	ret = -EINVAL;
	err("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
	dump_hex("<-- usb", data, len);
    }
    return(ret);
}

static int
send_next_queued(struct blackberry_device *blackberry_device, int mem_flags)
{
int ret = 0;
unsigned long flags;

    spin_lock_irqsave(&blackberry_device->urb_queue_lock, flags);
    {
	if (blackberry_device->last_urb)
	{
	    usb_free_urb(blackberry_device->last_urb);
	    blackberry_device->last_urb = NULL;
	}
	if (list_empty(&blackberry_device->urb_queue))
	{
	}
	else
	{
	struct urb *urb;

	    urb = list_entry(blackberry_device->urb_queue.next, struct urb, urb_list);
	    list_del(&urb->urb_list);
	    ret = usb_submit_urb(urb, mem_flags);
	    if (ret == 0)
	    {
		if (debug >= 1)
		{
		    info("%s(%d) - %08x", __FUNCTION__, __LINE__,
			blackberry_device->pin);
		    dump_hex("--> usb", urb->transfer_buffer, urb->transfer_buffer_length);
		}
		blackberry_device->last_urb = usb_get_urb(urb);
	    }
	    else
	    {
		usb_free_coherent(urb->dev,
		    urb->transfer_buffer_length, urb->transfer_buffer,
		    urb->transfer_dma);
		blackberry_put(blackberry_device);
	    }
	    usb_free_urb(urb);
	}
    }
    spin_unlock_irqrestore(&blackberry_device->urb_queue_lock, flags);
    return(ret);
}

static int
validate_socket_response(struct blackberry_mode *mode, const u_int8_t *data)
{
int ret = 0;
struct blackberry_device *blackberry_device = (mode && mode->tty) ? device_map[mode->tty->index].blackberry_device : NULL;

    switch (data[0])
    {
    case 0x11: /* auth failed */
	err("%s(%d) - %08x Bad password. %i more tries before something bad happens!!!", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    data[4]);
	if (blackberry_device)
	{
	    blackberry_device->attempts_remaining = data[4];
	}
	if (device_map[mode->tty->index].password_hash)
	{
	    kfree(device_map[mode->tty->index].password_hash);
	    device_map[mode->tty->index].password_hash = NULL;
	} /* fall through to try again */
    case 0x0e: /* auth challenge */
	{
	    if (blackberry_sha1_avail())
	    {
	    int use_old;

		if (device_map[mode->tty->index].password_hash == NULL)
		{
		    device_map[mode->tty->index].password_hash = kmalloc(4 + blackberry_sha1_digestsize(), GFP_KERNEL);
		    use_old = 0;
		}
		else
		{
		    use_old = 1;
		}
		if (device_map[mode->tty->index].password_hash)
		{
		    device_map[mode->tty->index].password_hash[0] = data[8];
		    device_map[mode->tty->index].password_hash[1] = data[9];
		    device_map[mode->tty->index].password_hash[2] = data[10];
		    device_map[mode->tty->index].password_hash[3] = data[11];
		    blackberry_device->attempts_remaining = data[3];
		    if (use_old)
		    {
			send_password(mode, GFP_KERNEL);
		    }
		    else if (&blackberry_device->desktop == mode &&
			     blackberry_device->desktop.send_packet)
		    {
		    unsigned char cmd[] = {0x40, 0x00, 'P'};

			/* should be a mode specific get_password call */
			(*blackberry_device->desktop.send_packet)(&blackberry_device->desktop, cmd, sizeof(cmd));
		    }
		    else
		    {
			err("%s(%d) - %08x Do not know how to get a password for \"%s\" mode",
			    __FUNCTION__, __LINE__,
			    blackberry_device->pin,
			    mode->mode);
		    }
		    mode->connected = 1;
		}
	    }
	    else
	    {
		err("%s(%d) - %08x", __FUNCTION__, __LINE__,
		    blackberry_device->pin);
	    }
	}
	break;
    case 0x10:
	mode->connected = 1;
#if 0
	if (&blackberry_device->desktop == mode)
	{
	unsigned char special[] = { 0x40, 0x00 };

	    /*
	    info("%s(%d) - %08x %p list_empty = %s", __FUNCTION__, __LINE__,
		blackberry_device->pin,
		blackberry_device->desktop.last_urb,
		list_empty(&blackberry_device->desktop.urb_queue) ? "True" : "False");
		*/
	    blackberry_write(blackberry_device, blackberry_device->desktop.socket, special, sizeof(special), GFP_KERNEL);
	}
#endif
	break;
    default:
	mode->connected = 0;
	break;
    }
    if (debug >= 8)
    {
	info("%s(%d) - %08x %s socket 0x%04x %s", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (uint32_t)-1,
	    mode ? mode->mode : "NULL",
	    mode ? mode->socket : -1,
	    mode && mode->connected ? "opened" : "closed");
    }
    if (mode->connected)
    {
	mode->msr |= TIOCM_DSR;
	mode->msr |= TIOCM_CD;
    }
    else
    {
	mode->socket = -1;
	mode->msr &= ~TIOCM_DSR;
	mode->msr &= ~TIOCM_CD;
    }
    wake_up_interruptible_sync(&mode->open_wait);
    return(ret);
}

static int
parse_socket_response(struct blackberry_device *blackberry_device, int len, const u_int8_t *data)
{
int ret = 0;

    if (debug >= 9)
    {
	info("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
    }

    if (len > 3)
    {
    int socket = 0;

	socket = (socket << 8) + data[2];
	socket = (socket << 8) + data[1];
	if (socket == blackberry_device->desktop.socket)
	{
	    send_next_queued(blackberry_device, GFP_KERNEL);
	    validate_socket_response(&blackberry_device->desktop, data);
	}
	else if (socket == blackberry_device->loader.socket)
	{
	    send_next_queued(blackberry_device, GFP_KERNEL);
	    validate_socket_response(&blackberry_device->loader, data);
	}
	else if (socket == blackberry_device->ftp.socket)
	{
	    send_next_queued(blackberry_device, GFP_KERNEL);
	    validate_socket_response(&blackberry_device->ftp, data);
	}
	else if (socket == blackberry_device->caltask.socket)
	{
	    send_next_queued(blackberry_device, GFP_KERNEL);
	    validate_socket_response(&blackberry_device->caltask, data);
	}
	else if (socket == blackberry_device->jvmdebug.socket)
	{
	    send_next_queued(blackberry_device, GFP_KERNEL);
	    validate_socket_response(&blackberry_device->jvmdebug, data);
	}
	else if (socket == blackberry_device->bootload.socket)
	{
	    send_next_queued(blackberry_device, GFP_KERNEL);
	    validate_socket_response(&blackberry_device->bootload, data);
	}
	else
	{
	    ret = -ENOSYS;
	    err("%s(%d) - %08x Unknown socket 0x%04x", __FUNCTION__, __LINE__,
		blackberry_device->pin, socket);
	    dump_hex("<-- usb", data, len);
	}
    }
    else
    {
	ret = -EINVAL;
	err("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
	dump_hex("<-- usb", data, len);
    }
    return(ret);
}

static int
parse_device_info(struct blackberry_device *blackberry_device, int len, const u_int8_t *data)
{
int ret = 0;

    if (debug >= 9)
    {
	info("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
    }

    if (len >= 6)
    {
    int type = 0;

	type = (type << 8) + data[6];
	type = (type << 8) + data[5];
	switch (type)
	{
	case 0x02:
	    if (len >= 24)
	    {
		blackberry_device->hwid = (blackberry_device->hwid << 8) + data[22];
		blackberry_device->hwid = (blackberry_device->hwid << 8) + data[21];
		blackberry_device->hwid = (blackberry_device->hwid << 8) + data[20];
		blackberry_device->hwid = (blackberry_device->hwid << 8) + data[19];
		info("%s(%d) - %08x \"%s\" %08x", __FUNCTION__, __LINE__,
		    blackberry_device->pin, &data[23], blackberry_device->hwid);
	    }
	    else
	    {
		err("%s(%d) - %08x", __FUNCTION__, __LINE__,
		    blackberry_device->pin);
	    }
	    break;
	case 0x04:
	    ret = parse_pin(blackberry_device, len - 7, &data[7]);
	    if (ret == 0)
	    {
		blackberry_device->max_packet = 0;
		blackberry_device->max_packet = (blackberry_device->max_packet << 8) + data[5];
		blackberry_device->max_packet = (blackberry_device->max_packet << 8) + data[4];
		/*
		blackberry_device->max_packet = MAX_PACKET;
		*/
		get_description(blackberry_device);
	    }
	    break;
	default:
	    ret = -ENOSYS;
	    err("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin, len);
	    dump_hex("<-- usb", data, len);
	    break;
	}
    }
    else
    {
	ret = -EINVAL;
	err("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
	dump_hex("<-- usb", data, len);
    }
    return(ret);
}

static int
process_input(struct blackberry_device *blackberry_device, int socket, int len, const u_int8_t *data)
{
int ret = 0;

    if (debug >= 8)
    {
	info("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, socket, len);
    }

    if (socket == 0)
    {
	switch (data[0])
	{
	case 0x02: /* hello response to 0x01 */
	    get_pin(blackberry_device);
	    break;
#if 0
	case 0x04: /* reset response to 0x03 */
	    break;
#endif
	case 0x06: /* info response to 0x05 */
	    ret = parse_device_info(blackberry_device, len - 1, &data[1]);
	    break;
	case 0x08: /* mode selected response to 0x07 */
	case 0x09: /* mode select failed response to 0x07 */
	    ret = parse_mode_select(blackberry_device, len - 1, &data[1]);
	    break;
	case 0x0e: /* auth challenge */
	case 0x11: /* auth failed */
	case 0x10: /* socket opened response to 0x0a */
	case 0x0c: /* socket closed response to 0x0b */
	    ret = parse_socket_response(blackberry_device, len, &data[0]);
	    break;
	case 0x13: /* sequence packet */
	    if (len >= 8)
	    {
	    int sequence = 0;

		sequence = (sequence << 8) + data[5];
		sequence = (sequence << 8) + data[4];
		if (data[1] == blackberry_device->desktop.socket)
		{
		unsigned char ack[] = { 0x06 };

		    tty_send(blackberry_device->desktop.tty, ack, sizeof(ack));
		    send_next_queued(blackberry_device, GFP_ATOMIC);
		}
		else if (data[1] == blackberry_device->loader.socket)
		{
		    send_next_queued(blackberry_device, GFP_ATOMIC);
		}
		else if (data[1] == blackberry_device->ftp.socket)
		{
		    send_next_queued(blackberry_device, GFP_ATOMIC);
		}
		else if (data[1] == blackberry_device->caltask.socket)
		{
		    send_next_queued(blackberry_device, GFP_ATOMIC);
		}
		else if (data[1] == blackberry_device->jvmdebug.socket)
		{
		    send_next_queued(blackberry_device, GFP_ATOMIC);
		}
		else if (data[1] == blackberry_device->bootload.socket)
		{
		    send_next_queued(blackberry_device, GFP_ATOMIC);
		}
		else
		{
		    ret = -EINVAL;
		    err("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
			blackberry_device->pin, socket, len);
		    dump_hex("<-- usb", data, len);
		}
	    }
	    else
	    {
		ret = -EINVAL;
		err("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
		    blackberry_device->pin, socket, len);
		dump_hex("<-- usb", data, len);
	    }
	    break;
	default:
	    ret = -ENOSYS;
	    err("%s(%d) - %08x socket = 0x%04x len = %i data[0] = 0x%02x", __FUNCTION__, __LINE__,
		blackberry_device->pin, socket, len, data[0]);
	    dump_hex("<-- usb", data, len);
	    break;
	}
    }
    else if (socket == blackberry_device->desktop.socket &&
             blackberry_device->desktop.send_packet)
    {
	if (debug >= 8)
	{
	    info("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin, socket, len);
	}
	(blackberry_device->desktop.send_packet)(&blackberry_device->desktop, data, len);
    }
    else if (socket == blackberry_device->loader.socket &&
             blackberry_device->loader.send_packet)
    {
	if (debug >= 8)
	{
	    info("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin, socket, len);
	    dump_hex("--> tty", data, len);
	}
	(blackberry_device->loader.send_packet)(&blackberry_device->loader, data, len);
    }
    else if (socket == blackberry_device->ftp.socket &&
             blackberry_device->ftp.send_packet)
    {
	if (debug >= 8)
	{
	    info("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin, socket, len);
	    dump_hex("--> tty", data, len);
	}
	(blackberry_device->ftp.send_packet)(&blackberry_device->ftp, data, len);
    }
    else if (socket == blackberry_device->caltask.socket &&
             blackberry_device->caltask.send_packet)
    {
	if (debug >= 8)
	{
	    info("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin, socket, len);
	    dump_hex("--> tty", data, len);
	}
	(blackberry_device->caltask.send_packet)(&blackberry_device->caltask, data, len);
    }
    else if (socket == blackberry_device->jvmdebug.socket &&
             blackberry_device->jvmdebug.send_packet)
    {
	if (debug >= 8)
	{
	    info("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin, socket, len);
	    dump_hex("--> tty", data, len);
	}
	(blackberry_device->jvmdebug.send_packet)(&blackberry_device->jvmdebug, data, len);
    }
    else if (socket == blackberry_device->bootload.socket &&
             blackberry_device->bootload.send_packet)
    {
	if (debug >= 8)
	{
	    info("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin, socket, len);
	    dump_hex("--> tty", data, len);
	}
	(blackberry_device->bootload.send_packet)(&blackberry_device->bootload, data, len);
    }
    else
    {
	ret = -ENOTSOCK;
	err("%s(%d) - %08x socket = 0x%04x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, socket, len);
	dump_hex("<-- usb", data, len);
    }
    return(ret);
}

static void
#ifdef HAVE_WORK_FUNC_T
blackberry_read_work(struct work_struct *work)
#else
blackberry_read_work(void *data)
#endif
{
int ret = 0;
#ifdef HAVE_WORK_FUNC_T
#else
struct work_struct *work = data;
#endif
struct blackberry_device *blackberry_device = container_of(work, struct blackberry_device, read_work);
struct urb *urb = blackberry_device->rx_urb;

    if (debug >= 9)
    {
	info("%s(%d) - %08x status = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, urb->status);
    }

    if (urb->status == 0)
    {
	if (urb->actual_length > 4)
	{
	u_int8_t *rx_data = urb->transfer_buffer;
	int socket = 0;
	int len = 0;

	    if (debug >= 1)
	    {
		info("%s(%d) - %08x actual_length = %i", __FUNCTION__, __LINE__,
		    blackberry_device->pin, urb->actual_length);
		dump_hex("<-- usb", urb->transfer_buffer, urb->actual_length);
	    }
	    socket = (socket << 8) + rx_data[1];
	    socket = (socket << 8) + rx_data[0];
	    len = (len << 8) + rx_data[3];
	    len = (len << 8) + rx_data[2];
	    if (len <= urb->actual_length)
	    {
#if 0
		ret = process_input(blackberry_device, socket, len - 4, &rx_data[4]);
#else
		ret = process_input(blackberry_device, socket, urb->actual_length - 4, &rx_data[4]);
#endif
		if (ret)
		{
		    err("%s(%d) - %08x actual_length = %i len = %i ret = %i", __FUNCTION__, __LINE__,
			blackberry_device->pin, urb->actual_length, len, ret);
		    dump_hex("<-- usb", urb->transfer_buffer, urb->actual_length);
		}
	    }
	    else
	    {
		err("%s(%d) - %08x actual_length = %i len = %i", __FUNCTION__, __LINE__,
		    blackberry_device->pin, urb->actual_length, len);
		dump_hex("<-- usb", urb->transfer_buffer, urb->actual_length);
	    }
	}
	else
	{
	    err("%s(%d) - %08x actual_length = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin, urb->actual_length);
	    dump_hex("<-- usb", urb->transfer_buffer, urb->actual_length);
	}
	ret = blackberry_read(blackberry_device);
    }
    else
    {
	err("%s(%d) - %08x status = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, urb->status);
    }

    blackberry_put(blackberry_device);
}

static void
#ifdef USB_COMPLETE_NEEDS_REGS
blackberry_read_callback(struct urb *urb, struct pt_regs *pt_regs)
#else
blackberry_read_callback(struct urb *urb)
#endif
{
int ret = 0;
struct blackberry_device *blackberry_device = urb->context;

    if (debug >= 9)
    {
	info("%s(%d) - %08x status = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, urb->status);
    }

#ifdef HAVE_WORK_FUNC_T
    PREPARE_WORK(&blackberry_device->read_work, blackberry_read_work);
#else
    PREPARE_WORK(&blackberry_device->read_work, blackberry_read_work, &blackberry_device->read_work);
#endif
    ret = queue_work(work_queue, &blackberry_device->read_work);
}

static int
blackberry_read(struct blackberry_device *blackberry_device)
{
int ret;

    if (debug >= 9)
    {
	info("%s(%d) - %08x max_packet = 0x%04x", __FUNCTION__, __LINE__,
	    blackberry_device->pin, blackberry_device->max_packet);
    }

    blackberry_get(blackberry_device);
    usb_fill_bulk_urb(blackberry_device->rx_urb,
	blackberry_device->usb_device,
	usb_rcvbulkpipe(blackberry_device->usb_device, blackberry_device->read_ep),
	blackberry_device->rx_buf,
	blackberry_device->rx_urb->transfer_buffer_length,
	blackberry_read_callback,
	blackberry_device);
    blackberry_device->rx_urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
    ret = usb_submit_urb(blackberry_device->rx_urb, GFP_KERNEL);
    if (ret == 0)
    {
    }
    else
    {
	err("%s(%d) - %08x ret = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, ret);
	blackberry_put(blackberry_device);
    }
    return(ret);
}

static void
blackberry_delete(struct kref *kref)
{
struct blackberry_device *blackberry_device = to_blackberry_device(kref);

    if (debug >= 0)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device->pin);
    }

#if 0
    /* Do I really need this?? */
    /* Don't think so */
    /* 6200 seems to need it */
    /* 8700 seems okay without it */
    /* 8800 seems okay without it */
    /* The reset causes problems on kernel 2.6.9
       It starts spitting out usb_storage errors, and
       rmmod'ing the blackberry module hangs rmmod :(
     */
    //if (blackberry_device->interface && blackberry_device->interface->cur_altsetting->desc.bNumEndpoints < 6)
    {
	usb_reset_device(blackberry_device->usb_device);
    }
#endif
    usb_put_dev(blackberry_device->usb_device);
    usb_free_coherent(blackberry_device->rx_urb->dev,
	blackberry_device->rx_urb->transfer_buffer_length, blackberry_device->rx_urb->transfer_buffer,
	blackberry_device->rx_urb->transfer_dma);
    usb_free_urb(blackberry_device->rx_urb);
    kfree(blackberry_device);
}

void
blackberry_put(struct blackberry_device *blackberry_device)
{
    if (debug >= 7)
    {
	info("%s(%d) - %08x refcount = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin,
	    atomic_read(&blackberry_device->kref.refcount));
    }
    kref_put(&blackberry_device->kref, blackberry_delete);
}

struct blackberry_device *
blackberry_get(struct blackberry_device *blackberry_device)
{
struct blackberry_device *ret = NULL;

    if (blackberry_device && atomic_read(&blackberry_device->kref.refcount) > 0)
    {
	kref_get(&blackberry_device->kref);
	ret = blackberry_device;
	if (debug >= 7)
	{
	    info("%s(%d) - %08x refcount = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin,
		atomic_read(&blackberry_device->kref.refcount));
	}
    }
    return(ret);
}

static int
blackberry_register(struct usb_driver *blackberry_driver)
{
int ret = 0;

    if (debug >= 8)
    {
	info("%s(%d)", __FUNCTION__, __LINE__);
    }

    if (ret == 0 && (ret = desktop_register()))
    {
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	goto bail_5;
    }
    if (ret == 0 && (ret = modem_register()))
    {
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	goto bail_4;
    }
    if (ret == 0 && (ret = loader_register()))
    {
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	goto bail_3;
    }
    if (ret == 0 && (ret = ftp_register()))
    {
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	goto bail_2;
    }
    if (ret == 0 && (ret = caltask_register()))
    {
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	goto bail_1;
    }
    if (ret == 0 && (ret = jvmdebug_register()))
    {
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	goto bail_0;
    }
    if (ret == 0 && (ret = bootload_register()))
    {
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	goto bail_7;
    }
    if (ret == 0 && (ret = usb_register(blackberry_driver)))
    {
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	goto bail_6;
    }
    return(ret);

bail_7:
    bootload_deregister();
bail_6:
    jvmdebug_deregister();
bail_0:
    caltask_deregister();
bail_1:
    ftp_deregister();
bail_2:
    loader_deregister();
bail_3:
    modem_deregister();
bail_4:
    desktop_deregister();
bail_5:
    return(ret);
}

static void
blackberry_deregister(struct usb_driver *blackberry_driver)
{
int i;

    if (debug >= 8)
    {
	info("%s(%d)", __FUNCTION__, __LINE__);
    }

    usb_deregister(blackberry_driver);
    for (i = 0; i < num_devices; i++)
    {
    	if (device_map[i].blackberry_device)
    	{
    	struct blackberry_device *blackberry_device;

	    blackberry_device = device_map[i].blackberry_device;
	    blackberry_unregister_device(blackberry_device);
	    device_map[i].blackberry_device = NULL;
	    blackberry_put(blackberry_device);
	}
    }
    bootload_deregister();
    jvmdebug_deregister();
    caltask_deregister();
    ftp_deregister();
    loader_deregister();
    modem_deregister();
    desktop_deregister();
}

static int __init
blackberry_init(void)
{
int ret = 0;
int i;

    if (debug >= 8)
    {
	info("%s(%d) - num_devices = %i debug = %i", __FUNCTION__, __LINE__,
	    num_devices, debug);
    }

    if (num_devices > 127)
    {
	err("%s(%d) - num_devices = %i, reset to 127.", __FUNCTION__, __LINE__, num_devices);
	num_devices = 127;
    }
    if (num_devices < 1 )
    {
	err("%s(%d) - num_devices = %i, reset to 1.", __FUNCTION__, __LINE__, num_devices);
	num_devices = 1;
    }

    if (ret == 0 && (device_map = kmalloc(num_devices * sizeof(struct device_map), GFP_KERNEL)) == NULL)
    {
    	ret = -ENOMEM;
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
    }
    if (ret == 0)
    {
	for (i = 0; i < num_devices; i++)
	{
	    device_map[i].pin = (u_int32_t)-1;
	    device_map[i].blackberry_device = NULL;
	    device_map[i].password_hash = NULL;
	    device_map[i].attempt_bootload = 0;
	}
    }
    if (ret == 0 && (work_queue = create_singlethread_workqueue(blackberry_driver.name)) == NULL)
    {
    	ret = -ENOMEM;
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	kfree(device_map);
	device_map = NULL;
    }
    if (ret == 0)
    {
	blackberry_sha1_init();
	if (blackberry_sha1_avail())
	{
	    if (debug > 0)
	    {
		info("%s(%d) - passwords are possible", __FUNCTION__, __LINE__);
	    }
	}
	else
    	{
	    err("%s(%d) - passwords not supported", __FUNCTION__, __LINE__);
    	}
    }
    if (ret == 0 && (ret = blackberry_register(&blackberry_driver)))
    {
	err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	blackberry_sha1_dnit();
	destroy_workqueue(work_queue);
	work_queue = NULL;
	kfree(device_map);
	device_map = NULL;
    }
    return(ret);
}

static void __exit
blackberry_exit(void)
{
    if (debug >= 8)
    {
	info("%s(%d) - num_devices = %i", __FUNCTION__, __LINE__, num_devices);
    }

    blackberry_sha1_dnit();

    blackberry_deregister(&blackberry_driver);
    if (work_queue)
    {
	destroy_workqueue(work_queue);
	work_queue = NULL;
    }
    if (device_map)
    {
    int i;

	for (i = 0; i < num_devices; i++)
	{
	    if (device_map[i].password_hash)
	    {
		kfree(device_map[i].password_hash);
	    }
	}
	kfree(device_map);
	device_map = NULL;
    }
}

static int
probe(struct usb_interface *intf, const struct usb_device_id *id)
{
int ret = 0;
struct usb_device *udev = interface_to_usbdev(intf);

    if (debug >= 1)
    {
	info("%s(%d) - bcdDevice 0x%04x bMaxPower %i class 0x%02x sub-class 0x%02x endpoints %i protocol 0x%02x", __FUNCTION__, __LINE__,
	    udev->descriptor.bcdDevice,
	    udev->config->desc.bMaxPower,
	    intf->cur_altsetting->desc.bInterfaceClass,
	    intf->cur_altsetting->desc.bInterfaceSubClass,
	    intf->cur_altsetting->desc.bNumEndpoints,
	    intf->cur_altsetting->desc.bInterfaceProtocol);
    }
    if (intf->cur_altsetting->desc.bInterfaceClass == 0xff)
    {
	if (udev->config->desc.bMaxPower < 250)
	{
	    ret = set_charge(intf);
	}
	if (ret == 0)
	{
	    if (intf->cur_altsetting->desc.bNumEndpoints >= 4)
	    {
	    struct blackberry_device *blackberry_device = NULL;

		if (debug >= 8)
		{
		    info("%s(%d) - 0x%02x 0x%02x", __FUNCTION__, __LINE__,
			intf->cur_altsetting->endpoint[2].desc.bEndpointAddress,
			intf->cur_altsetting->endpoint[3].desc.bEndpointAddress);
	    	}
	    	blackberry_device = kmalloc(sizeof(*blackberry_device), GFP_KERNEL);
	    	if (blackberry_device)
	    	{
		    memset(blackberry_device, 0, sizeof(*blackberry_device));
		    blackberry_device->rx_urb = usb_alloc_urb(0, GFP_KERNEL);
		    if (blackberry_device->rx_urb)
		    {
			blackberry_device->usb_device = usb_get_dev(udev);
			blackberry_device->rx_urb->transfer_buffer_length = 0x4000;
			blackberry_device->rx_buf = usb_alloc_coherent(blackberry_device->usb_device, blackberry_device->rx_urb->transfer_buffer_length, GFP_KERNEL, &blackberry_device->rx_urb->transfer_dma);
			if (blackberry_device->rx_buf)
			{
			    kref_init(&blackberry_device->kref);
			    blackberry_device->interface = intf;
			    blackberry_device->pin = (u_int32_t)-1;
			    blackberry_device->hwid = (u_int32_t)-1;
			    blackberry_device->max_packet = 32;
			    blackberry_device->read_ep = intf->cur_altsetting->endpoint[2].desc.bEndpointAddress;
			    blackberry_device->write_ep = intf->cur_altsetting->endpoint[3].desc.bEndpointAddress;
#ifdef HAVE_WORK_FUNC_T
			    INIT_WORK(&blackberry_device->read_work, blackberry_read_work);
#else
			    INIT_WORK(&blackberry_device->read_work, blackberry_read_work, NULL);
#endif
			    INIT_LIST_HEAD(&blackberry_device->urb_queue);
			    spin_lock_init(&blackberry_device->urb_queue_lock);
			    blackberry_device->last_urb = NULL;

			    usb_set_intfdata(blackberry_device->interface, blackberry_device);

#if 1
			    if (blackberry_device->usb_device->descriptor.bcdDevice < 0x0100)
			    {
				usb_reset_device(blackberry_device->usb_device);
			    }
#endif
			    ret = blackberry_read(blackberry_device);
			    if (ret == 0)
			    {
			    char hello[] = { 0x01, 0xff, 0x00 };
			    int written;

				written = blackberry_write(blackberry_device, 0, hello, sizeof(hello), GFP_KERNEL);
				if (written == sizeof(hello))
				{
				    ret = 0;
				}
				else
				{
				    err("%s(%d)", __FUNCTION__, __LINE__);
				    ret = -ENODEV;
				}
			    }
			    else
			    {
				err("%s(%d) - ret = %i", __FUNCTION__, __LINE__, ret);
			    }
			    if (ret != 0)
			    {
				err("%s(%d)", __FUNCTION__, __LINE__);
				blackberry_put(blackberry_device);
				ret = -ENODEV;
			    }
			}
			else
			{
			    usb_put_dev(blackberry_device->usb_device);
			    usb_free_urb(blackberry_device->rx_urb);
			    kfree(blackberry_device);
			    err("%s(%d) - Out of memory", __FUNCTION__, __LINE__);
			    ret = -ENOMEM;
			}
		    }
		    else
		    {
			kfree(blackberry_device);
			err("%s(%d) - Out of memory", __FUNCTION__, __LINE__);
			ret = -ENOMEM;
		    }
	    	}
	    	else
	    	{
		    err("%s(%d) - Out of memory", __FUNCTION__, __LINE__);
		    ret = -ENOMEM;
	    	}
	    }
	    else
	    {
		err("%s(%d) - Only %i endpoints, can't continue", __FUNCTION__, __LINE__,
		    intf->cur_altsetting->desc.bNumEndpoints);
		ret = -ENODEV;
	    }
	}
	else
	{
	    info("%s(%d) - Charge current adjusted", __FUNCTION__, __LINE__);
	}
    }
    else if (udev->descriptor.idProduct == 0x06 && intf->cur_altsetting->desc.bInterfaceClass == 0x08)
    {
    char *buf;

    	buf = kmalloc(2, GFP_KERNEL);
    	ret = usb_control_msg(udev, usb_rcvctrlpipe(udev, 0),
	    0xa9,
	    USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
	    1,
	    1,
	    buf,
	    2,
	    USB_CTRL_SET_TIMEOUT);
	kfree(buf);
	if (ret >= 0)
    	{
	    info("%s(%d) - %i", __FUNCTION__, __LINE__, ret);
	    if (udev->config->desc.bMaxPower < 250)
	    {
		ret = set_charge(intf);
	    }
	    else
	    {
		ret = usb_reset_configuration(udev);
	    }
	    if (ret == 0)
	    {
		ret = usb_reset_configuration(udev);
	    }
    	}
    	else
	{
	    info("%s(%d)", __FUNCTION__, __LINE__);
	    //ret = usb_reset_configuration(udev);
	}

    	if (ret)
    	{
	    info("%s(%d)", __FUNCTION__, __LINE__);
    	}
    	else
    	{
	    ret = -ENODEV;
    	}
	ret = -ENODEV;
    }
    else
    {
	err("%s(%d) - class 0x%02x looking for 0xff", __FUNCTION__, __LINE__,
	    intf->cur_altsetting->desc.bInterfaceClass);
	ret = -ENODEV;
    }
    //info("%s(%d) - ret = %i", __FUNCTION__, __LINE__, ret);
    return(ret);
}

static void
disconnect(struct usb_interface *intf)
{
struct blackberry_device *blackberry_device = usb_get_intfdata(intf);

    if (debug >= 8)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
    }

#ifdef HAVE_LOCK_KERNEL
    lock_kernel();
#endif
    blackberry_device = usb_get_intfdata(intf);
    usb_set_intfdata(blackberry_device->interface, NULL);
#ifdef HAVE_LOCK_KERNEL
    unlock_kernel();
#endif

    blackberry_unregister_device(blackberry_device);
    blackberry_put(blackberry_device);
}

module_init(blackberry_init);
module_exit(blackberry_exit);

MODULE_DEVICE_TABLE(usb, blackberry_table);
