/*
	XmBlackBerry, Copyright (C) 2008  Rick Scott <rwscott@users.sourceforge.net>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 2.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
    Copyright 2008, Chris Frey, NetDirect (http://www.netdirect.ca/), GPL v2
    2008/11 - Now compiles on 2.6.26.5: Changes to URB callback mechanism and
	      work queue API.  Also uses new sha1 API.
*/

#include "config.h"

#include <linux/usb.h>
#include <linux/tty.h>
#include <linux/tty_driver.h>
#include <linux/tty_flip.h>
#include <linux/mm.h>
#include <asm/uaccess.h>

#include "blackberry.h"
#include "modem.h"
#include "util.h"
#include "crypto.h"

static struct tty_driver *modem_driver = NULL;
static char special_tag[] = {0x78, 0x56, 0x34, 0x12};
static char special_sauce[] = {
    0x01, 0x00, 0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x78, 0x56, 0x34, 0x12};
static char to_command[] = {
    0x00, 0x00, 0x00, 0x00, 0xb0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xc2, 0x01, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x56, 0x34, 0x12};
static char disconnect[] = {
    0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x56, 0x34, 0x12};
/*
static char unknown[] = {
    0x00, 0x00, 0x00, 0x00, 0x30, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0xc2, 0x01, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x56, 0x34, 0x12};
*/
static char stop[] = {
    0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x56, 0x34, 0x12};

static int modem_read(struct blackberry_device *blackberry_device);

static int modem_open(struct tty_struct *tty, struct file *file);
static void modem_close(struct tty_struct *tty, struct file *file);
#ifdef FROM_USER_IN_TTY_OPS
static int modem_write(struct tty_struct *tty, int from_user, const unsigned char *buf, int count);
#else
static int modem_write(struct tty_struct *tty, const unsigned char *buf, int count);
#endif
static int modem_write_room(struct tty_struct *tty);
static int modem_chars_in_buffer(struct tty_struct *tty);
static struct tty_operations modem_ops = {
    .open = modem_open,
    .close = modem_close,
    .write = modem_write,
    .write_room = modem_write_room,
    .chars_in_buffer = modem_chars_in_buffer,
};

static void
#ifdef USB_COMPLETE_NEEDS_REGS
modem_write_callback(struct urb *urb, struct pt_regs *regs)
#else
modem_write_callback(struct urb *urb)
#endif
{
struct blackberry_device *blackberry_device = urb->context;

    if (debug >= 9)
    {
	info("%s(%d) - %08x status = %i", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    urb->status);
    }

    usb_free_coherent(urb->dev,
    	urb->transfer_buffer_length, urb->transfer_buffer,
    	urb->transfer_dma);
    blackberry_put(blackberry_device);
}

static int
modem_blackberry_write(struct blackberry_device *blackberry_device, int from_user, const void *input, int len, int mem_flags)
{
int written = 0;

    if (debug >= 9)
    {
	info("%s(%d) - %08x len %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, len);
    }

    if (len > 0)
    {
    struct urb *urb;

	urb = usb_alloc_urb(0, mem_flags);
	if (urb)
	{
	char *data;

	    data = usb_alloc_coherent(blackberry_device->usb_device, len, mem_flags, &urb->transfer_dma);
	    if (data)
	    {
	    int ret;

		if (from_user)
		{
		    ret = copy_from_user(&data[0], input, len);
		}
		else
		{
		    memcpy(&data[0], input, len);
		}

		blackberry_get(blackberry_device);
		usb_fill_bulk_urb(urb, blackberry_device->usb_device,
		    usb_sndbulkpipe(blackberry_device->usb_device, blackberry_device->modem.write_ep),
		    data, len,
		    modem_write_callback, blackberry_device);
		urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
		ret = usb_submit_urb(urb, GFP_ATOMIC);
		if (ret == 0)
		{
		    if (debug >= 1)
		    {
			info("%s(%d) - %08x", __FUNCTION__, __LINE__,
			    blackberry_device->pin);
			dump_hex("--> usb modem", urb->transfer_buffer, urb->transfer_buffer_length);
		    }
		    blackberry_device->modem.tx_urbs++;
		    written = len;
		}
		else
		{
		    written = ret;
		    err("%s(%d) - %08x ret = %i", __FUNCTION__, __LINE__,
			blackberry_device->pin,
			ret);
		    usb_free_coherent(blackberry_device->usb_device, urb->transfer_buffer_length, urb->transfer_buffer, urb->transfer_dma);
		    blackberry_put(blackberry_device);
		}
	    }
	    else
	    {
		written = -ENOMEM;
		err("%s(%d) - %08x ENOMEM", __FUNCTION__, __LINE__,
		    blackberry_device->pin);
	    }
	    usb_free_urb(urb);
	}
	else
	{
	    written = -ENOMEM;
	    err("%s(%d) - %08x ENOMEM", __FUNCTION__, __LINE__,
		blackberry_device->pin);
	}
    }
    else
    {
	err("%s(%d) - %08x len = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin,
	    len);
    }
    return(written);
}

static int
tty_write_str(struct blackberry_device *blackberry_device, const void *input)
{
    return(tty_send(blackberry_device->modem.tty, input, strlen(input)));
}

static void
#ifdef HAVE_WORK_FUNC_T
modem_read_work(struct work_struct *work)
#else
modem_read_work(void *data)
#endif
{
int ret = 0;
#ifdef HAVE_WORK_FUNC_T
#else
struct work_struct *work = data;
#endif
struct modem *modem = container_of(work, struct modem, read_work);
struct urb *urb = modem->rx_urb;
struct blackberry_device *blackberry_device = urb->context;
unsigned char *rx_data;

    if (debug >= 9)
    {
	info("%s(%d) - %08x status = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, urb->status);
    }


    if (urb->actual_length >= urb->transfer_buffer_length)
    {
	info("%s(%d) - %08x actual_length = %i buffer_length = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin,
	    urb->actual_length,
	    urb->transfer_buffer_length);
	blackberry_device->modem.extra_space++;
    }
    if (debug >= 1)
    {
	info("%s(%d) - %08x actual_length = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, urb->actual_length);
	dump_hex("<-- usb modem", urb->transfer_buffer, urb->actual_length);
    }
    rx_data = urb->transfer_buffer;
    blackberry_device->modem.rx_urbs++;

    if (urb->actual_length >= sizeof(special_tag) && memcmp(&rx_data[urb->actual_length - sizeof(special_tag)], &special_tag[0], sizeof(special_tag)) == 0)
    {
    unsigned char unknown1[] = {0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                0x01, 0x00, 0x00, 0x00, 0x1c, 0x00, 0x00, 0x00, 0x78, 0x56, 0x34, 0x12};
    unsigned char unknown2[] = {0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
                                0x18, 0x00, 0x00, 0x00, 0x78, 0x56, 0x34, 0x12};
    unsigned char unknown3[] = {0x02, 0x00, 0x00, 0x00, 0xcc, 0x12, 0x29, 0x27, 0x0a, 0x00, 0x00, 0x00, 0x0a, 0x00, 0x00, 0x00,
                                0x98, 0xcc, 0xf2, 0x23, 0x1c, 0x00, 0x00, 0x00,  0x78, 0x56, 0x34, 0x12};

	if      (urb->actual_length == sizeof(unknown1) && memcmp(urb->transfer_buffer, unknown1, 4) == 0)
	{
	/*
	unsigned char cmd[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0xc2, 0x01, 0x00,
	                       0x40, 0x7d, 0x20, 0x80, 0x7d, 0x26, 0x7d, 0x3d, 0x78, 0x56, 0x34, 0x12};
	*/
	unsigned char cmd[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x03, 0x00, 0x00, 0x00, 0x00, 0xc2, 0x01, 0x00,
	                       0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x78, 0x56, 0x34, 0x12};

	    if (rx_data[16] == 0x01)
	    {
	    }
	    else
	    {
		err("%s(%d) - %08x password not available", __FUNCTION__, __LINE__,
		    blackberry_device->pin);
		tty_write_str(blackberry_device, "ERROR INVALID PASSWORD\r\n");
	    }
	    modem_blackberry_write(blackberry_device, 0, cmd, sizeof(cmd), GFP_KERNEL);
	    blackberry_device->modem.login_complete = 1;
	    wake_up_interruptible_sync(&blackberry_device->modem.open_wait);
	}
	else if (urb->actual_length == sizeof(unknown2) && memcmp(urb->transfer_buffer, unknown2, sizeof(urb->actual_length)) == 0)
	{
	}
	else if (urb->actual_length == sizeof(unknown3) && memcmp(urb->transfer_buffer, unknown3, 4) == 0)
	{
	    /*
	    info("%s(%d) - %08x attempts_left = %i", __FUNCTION__, __LINE__,
		blackberry_device->pin, data[8]);
	    dump_hex("password request", urb->transfer_buffer, urb->actual_length);
	    */
	    if (device_map[blackberry_device->modem.tty->index].password_hash)
	    {
	    u8 *response = NULL;

		if (blackberry_sha1_avail())
		{
		    response = kmalloc(4 + blackberry_sha1_digestsize() + sizeof(special_tag), GFP_KERNEL);
		}
		if (response)
		{
		u8 *data = urb->transfer_buffer;
		u8 hdr[] = {0x03, 0x00, 0x00, 0x00};

		    memcpy(&device_map[blackberry_device->modem.tty->index].password_hash[0], &data[4], 4);

		    blackberry_sha1_sum(
			&device_map[blackberry_device->modem.tty->index].password_hash[0],
			4 + blackberry_sha1_digestsize(),
			&response[sizeof(hdr)]);

		    memcpy(&response[0], hdr, sizeof(hdr));
		    memcpy(&response[sizeof(hdr) + blackberry_sha1_digestsize()], special_tag, sizeof(special_tag));
		    modem_blackberry_write(blackberry_device, 0,
			response,
			4 + blackberry_sha1_digestsize() + sizeof(special_tag),
			GFP_KERNEL);

		    kfree(response);
		}
		else
		{
		    err("%s(%d) - %08x ENOMEM", __FUNCTION__, __LINE__,
			blackberry_device->pin);
		    tty_write_str(blackberry_device, "ERROR\r\n");
		    wake_up_interruptible_sync(&blackberry_device->modem.open_wait);
		}
	    }
	    else
	    {
		err("%s(%d) - %08x password not available", __FUNCTION__, __LINE__,
		    blackberry_device->pin);
		tty_write_str(blackberry_device, "ERROR NO PASSWORD\r\n");
		wake_up_interruptible_sync(&blackberry_device->modem.open_wait);
	    }
	}
	else
	{
	    err("%s(%d) - %08x", __FUNCTION__, __LINE__,
		blackberry_device->pin);
	    dump_hex("<-- usb modem special", urb->transfer_buffer, urb->actual_length);
	}
	blackberry_device->modem.rx_special_urbs++;
    }
    else
    {
	tty_send(blackberry_device->modem.tty, urb->transfer_buffer, urb->actual_length);
    }

    ret = modem_read(blackberry_device);
    blackberry_put(blackberry_device);
}

static void
#ifdef USB_COMPLETE_NEEDS_REGS
modem_read_callback(struct urb *urb, struct pt_regs *pt_regs)
#else
modem_read_callback(struct urb *urb)
#endif
{
int ret = 0;
struct blackberry_device *blackberry_device = urb->context;

    if (debug >= 9)
    {
	info("%s(%d) - %08x status = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, urb->status);
    }

    if (urb->actual_length > blackberry_device->modem.max_rx)
    {
	blackberry_device->modem.max_rx = urb->actual_length;
    }
    if (urb->status == 0 ||
        urb->status == -EOVERFLOW)
    {
#ifdef HAVE_WORK_FUNC_T
	PREPARE_WORK(&blackberry_device->modem.read_work, modem_read_work);
#else
	PREPARE_WORK(&blackberry_device->modem.read_work, modem_read_work, &blackberry_device->modem.read_work);
#endif
	ret = queue_work(work_queue, &blackberry_device->modem.read_work);
    }
    else
    {
	err("%s(%d) - %08x status = %i actual_length = 0x%04x transfer_buffer_length = 0x%04x", __FUNCTION__, __LINE__,
	    blackberry_device->pin,
	    urb->status,
	    urb->actual_length,
	    urb->transfer_buffer_length);
    }
}

static int
modem_read(struct blackberry_device *blackberry_device)
{
int ret;

    if (debug >= 9)
    {
	info("%s(%d) - %08x max_packet = 0x%04x", __FUNCTION__, __LINE__,
	    blackberry_device->pin, blackberry_device->max_packet);
    }

    blackberry_get(blackberry_device);
    usb_fill_bulk_urb(blackberry_device->modem.rx_urb,
	blackberry_device->usb_device,
	usb_rcvbulkpipe(blackberry_device->usb_device, blackberry_device->modem.read_ep),
	blackberry_device->modem.rx_buf,
	blackberry_device->modem.rx_urb->transfer_buffer_length,
	modem_read_callback,
	blackberry_device);
    blackberry_device->modem.rx_urb->transfer_flags |= URB_NO_TRANSFER_DMA_MAP;
    ret = usb_submit_urb(blackberry_device->modem.rx_urb, GFP_KERNEL);
    if (ret == 0)
    {
    }
    else
    {
	err("%s(%d) - %08x ret = %i", __FUNCTION__, __LINE__,
	    blackberry_device->pin, ret);
	blackberry_put(blackberry_device);
    }
    return(ret);
}

static int
modem_open(struct tty_struct *tty, struct file *file)
{
int ret = 0;
struct blackberry_device *blackberry_device = tty ? device_map[tty->index].blackberry_device : NULL;

    if (debug >= 7)
    {
	info("%s(%d) - %08x index = %i driver_data = 0x%p", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    tty ? tty->index : -1,
	    tty ? tty->driver_data : NULL);
    }

    if ((blackberry_device = blackberry_get(blackberry_device)))
    {
	if (down_interruptible(&blackberry_device->modem.sem))
	{
	    ret = -ERESTARTSYS;
	}
	else
	{
	    if (tty->driver_data == NULL)
	    {
		blackberry_get(blackberry_device);
		tty->driver_data = blackberry_device;
		blackberry_device->modem.tty = tty;
	    }
	    ++blackberry_device->modem.open_count;
	    if (blackberry_device->modem.open_count == 1)
	    {
		tty->low_latency = 1;
		blackberry_device->modem.tx_urbs = 0;
		blackberry_device->modem.rx_urbs = 0;
		blackberry_device->modem.rx_special_urbs = 0;
		ret = modem_read(blackberry_device);
		if (ret == 0)
		{
		int written;

		    written = modem_blackberry_write(blackberry_device, 0, special_sauce, sizeof(special_sauce), GFP_KERNEL);
		    if (written == sizeof(special_sauce))
		    {
			if (file->f_flags & O_NONBLOCK)
			{
			}
			else
			{
			    wait_event_interruptible(blackberry_device->modem.open_wait, (blackberry_device->modem.login_complete));
			    if (blackberry_device->modem.login_complete)
			    {
			    }
			    else
			    {
				ret = -ENODEV;
			    }
			}
		    }
		    else
		    {
			ret = -ENODEV;
			err("%s(%d) - %08x", __FUNCTION__, __LINE__,
			    blackberry_device ? blackberry_device->pin: (u_int32_t)-1);
			--blackberry_device->modem.open_count;
			tty->driver_data = NULL;
			blackberry_device->modem.tty = NULL;
			blackberry_put(blackberry_device);
		    }
		}
		else
		{
		    ret = -ENODEV;
		    err("%s(%d) - %08x", __FUNCTION__, __LINE__,
			blackberry_device ? blackberry_device->pin: (u_int32_t)-1);
		    --blackberry_device->modem.open_count;
		    tty->driver_data = NULL;
		    blackberry_device->modem.tty = NULL;
		    blackberry_put(blackberry_device);
		}
	    }
	    up(&blackberry_device->modem.sem);
	}
	blackberry_put(blackberry_device);
    }
    else
    {
	ret = -ENODEV;
	err("%s(%d) - %08x Modem not available", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin: (u_int32_t)-1);
    }
    return(ret);
}

static void
modem_close(struct tty_struct *tty, struct file *file)
{
struct blackberry_device *blackberry_device = tty ? tty->driver_data : NULL;

    if (debug >= 8)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
    }
    if ((blackberry_device = blackberry_get(blackberry_device)))
    {
	down(&blackberry_device->modem.sem);
	if (blackberry_device->modem.open_count > 0)
	{
	    --blackberry_device->modem.open_count;
	    if (blackberry_device->modem.open_count <= 0)
	    {
		modem_blackberry_write(blackberry_device, 0, to_command, sizeof(to_command), GFP_KERNEL);
		modem_blackberry_write(blackberry_device, 0, disconnect, sizeof(disconnect), GFP_KERNEL);
		modem_blackberry_write(blackberry_device, 0, stop, sizeof(stop), GFP_KERNEL);
		info("%s(%d) - %08x tx = %i rx = %i rx_special = %i rx_data = %i extra_space = %i max_rx = %i", __FUNCTION__, __LINE__,
		    blackberry_device->pin,
		    blackberry_device->modem.tx_urbs,
		    blackberry_device->modem.rx_urbs,
		    blackberry_device->modem.rx_special_urbs,
		    blackberry_device->modem.rx_urbs - blackberry_device->modem.rx_special_urbs,
		    blackberry_device->modem.extra_space,
		    blackberry_device->modem.max_rx);
		usb_kill_urb(blackberry_device->modem.rx_urb);
		tty->driver_data = NULL;
		blackberry_device->modem.tty = NULL;
		blackberry_put(blackberry_device);
	    }
	}
	else
	{
	    err("%s(%d) - %08x close without an open", __FUNCTION__, __LINE__,
		blackberry_device->pin);
	}
	up(&blackberry_device->modem.sem);
	blackberry_put(blackberry_device);
    }
    else
    {
	err("%s(%d) - %08x Modem not available", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
    }
}

static int
#ifdef FROM_USER_IN_TTY_OPS
modem_write(struct tty_struct *tty, int from_user, const unsigned char *buf, int count)
#else
modem_write(struct tty_struct *tty, const unsigned char *buf, int count)
#endif
{
int written = 0;
struct blackberry_device *blackberry_device = tty ? tty->driver_data : NULL;

    if (blackberry_device && blackberry_device->modem.open_count)
    {
	if (count > 0)
	{
#ifdef FROM_USER_IN_TTY_OPS
	    written = modem_blackberry_write(blackberry_device, from_user, buf, count, GFP_ATOMIC);
#else
	    written = modem_blackberry_write(blackberry_device, 0, buf, count, GFP_ATOMIC);
#endif
    	}
    }
    else
    {
	written = -ENODEV;
	err("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
    }
    return(written);
}

static int
modem_write_room(struct tty_struct *tty)
{
int ret = 0;
struct blackberry_device *blackberry_device = tty ? tty->driver_data : NULL;

    if (debug >= 2)
    {
	info("%s(%d) - %08x tty = %p max_packet = %i", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    tty,
	    blackberry_device ? blackberry_device->max_packet : -1);
    }

    if (blackberry_device)
    {
    	ret = blackberry_device->max_packet;
    }
    else
    {
	ret = -ENODEV;
	err("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
    }
    return(ret);
}

static int
modem_chars_in_buffer(struct tty_struct *tty)
{
int num_chars = 0;
struct blackberry_device *blackberry_device = tty ? tty->driver_data : NULL;

    if (debug >= 9)
    {
	info("%s(%d) - %08x tty = %p", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    tty);
    }

    return(blackberry_device ? num_chars : -ENODEV);
}

int
modem_register_device(struct modem *modem, int index)
{
int ret = 0;
struct blackberry_device *blackberry_device = container_of(modem, struct blackberry_device, modem);

    if (debug >= 9)
    {
	info("%s(%d) - %08x", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1);
    }
    blackberry_get(blackberry_device);
    if (blackberry_device->interface->cur_altsetting->desc.bNumEndpoints >= 6)
    {
	modem->rx_urb = usb_alloc_urb(0, GFP_KERNEL);
	if (modem->rx_urb)
	{
	    modem->rx_urb->transfer_buffer_length = 0x4000;
	    modem->rx_buf = usb_alloc_coherent(blackberry_device->usb_device, modem->rx_urb->transfer_buffer_length, GFP_KERNEL, &modem->rx_urb->transfer_dma);
	    if (modem->rx_buf)
	    {
		blackberry_get(blackberry_device);
		if (debug >= 8)
		{
		    info("%s(%d) - 0x%02x 0x%02x", __FUNCTION__, __LINE__,
			blackberry_device->interface->cur_altsetting->endpoint[4].desc.bEndpointAddress,
			blackberry_device->interface->cur_altsetting->endpoint[5].desc.bEndpointAddress);
		}
		modem->extra_space = 1533;
		modem->max_rx = 0;
		modem->open_count = 0;
		modem->tx_urbs = 0;
		modem->rx_urbs = 0;
		modem->rx_special_urbs = 0;
		modem->read_ep = blackberry_device->interface->cur_altsetting->endpoint[4].desc.bEndpointAddress;
		modem->write_ep = blackberry_device->interface->cur_altsetting->endpoint[5].desc.bEndpointAddress;
#ifdef HAVE_WORK_FUNC_T
		INIT_WORK(&modem->read_work, modem_read_work);
#else
		INIT_WORK(&modem->read_work, modem_read_work, NULL);
#endif
		sema_init(&modem->sem, 1);
		modem->login_complete = 0;
		init_waitqueue_head(&modem->open_wait);
		tty_register_device(modem_driver, index, NULL);
		if (debug >= 1)
		{
		    info("%s(%d) - %08x /dev/%s%i", __FUNCTION__, __LINE__,
			blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
			modem_driver->name, index);
		}
	    }
	    else
	    {
		ret = -ENOMEM;
	    }
	}
	else
	{
	    ret = -ENOMEM;
	}
    }
    else
    {
    	ret = -ENODEV;
    }
    blackberry_put(blackberry_device);
    return(ret);
}

int
modem_deregister_device(struct modem *modem, int index)
{
int ret = 0;
struct blackberry_device *blackberry_device = container_of(modem, struct blackberry_device, modem);

    if (debug >= 9)
    {
	info("%s(%d) - %08x /dev/%s%i", __FUNCTION__, __LINE__,
	    blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
	    modem_driver->name,
	    index);
    }
    blackberry_get(blackberry_device);
    if (blackberry_device->interface->cur_altsetting->desc.bNumEndpoints >= 6)
    {
	if (debug >= 1)
	{
	    info("%s(%d) - %08x /dev/%s%i", __FUNCTION__, __LINE__,
		blackberry_device ? blackberry_device->pin : (u_int32_t)-1,
		modem_driver->name, index);
	}

	modem->login_complete = 1;
	wake_up_interruptible_sync(&modem->open_wait);
	usb_free_coherent(modem->rx_urb->dev,
	    modem->rx_urb->transfer_buffer_length, modem->rx_urb->transfer_buffer,
	    modem->rx_urb->transfer_dma);
	usb_free_urb(modem->rx_urb);
	tty_unregister_device(modem_driver, index);
	blackberry_put(blackberry_device);
    }
    else
    {
    	ret = -ENODEV;
    }
    blackberry_put(blackberry_device);
    return(ret);
}

int
modem_register(void)
{
int ret = 0;

    if (debug >= 9)
    {
	info("%s(%d)", __FUNCTION__, __LINE__);
    }

    modem_driver = alloc_tty_driver(num_devices);
    if (modem_driver)
    {
    	modem_driver->owner = THIS_MODULE;
    	modem_driver->driver_name = "Blackberry Modem";
    	modem_driver->name = "bbmodem";
    	modem_driver->flags = TTY_DRIVER_RESET_TERMIOS | TTY_DRIVER_REAL_RAW
#ifdef TTY_DRIVER_NO_DEVFS
    	| TTY_DRIVER_NO_DEVFS;
#else
    	| TTY_DRIVER_DYNAMIC_DEV;
#endif
    	modem_driver->init_termios = tty_std_termios;
    	tty_set_operations(modem_driver, &modem_ops);
    	ret = tty_register_driver(modem_driver);
    	if (ret)
    	{
	    err("%s(%d) - ret = %d", __FUNCTION__, __LINE__, ret);
	    put_tty_driver(modem_driver);
	    modem_driver = NULL;
    	}
    }
    else
    {
    	ret = -ENOMEM;
	err("%s(%d)", __FUNCTION__, __LINE__);
    }
    return(ret);
}

int
modem_deregister(void)
{
int ret = 0;

    if (debug >= 9)
    {
	info("%s(%d)", __FUNCTION__, __LINE__);
    }
    tty_unregister_driver(modem_driver);
    put_tty_driver(modem_driver);
    modem_driver = NULL;
    return(ret);
}
